/*
	HEADER:		CUG267;
	TITLE:		8085 Cross-Assembler (Portable);
	FILENAME:	was A85UTIL.C; now A8008UTIL.C
	VERSION:	0.1;
	DATE:		08/27/1988;
	SEE-ALSO:	A8008.H;
	AUTHORS:	William C. Colley III;
*/

/*
		      8085 Cross-Assembler in Portable C

		Copyright (c) 1985,1987 William C. Colley, III

Revision History:

Ver	Date		Description

0.0	AUG 1987	A85: Derived from version 3.4 of my portable 6800/6801
			cross-assembler.  WCC3.

0.1	AUG 1988	A85: Fixed a bug in the command line parser that puts it
			into a VERY long loop if the user types a command line
			like "A85 FILE.ASM -L".  WCC3 per Alex Cameron.


	DEC 2013	A85: changes by Herb Johnson HRJ to compile under lcc-32 for MS-DOS.
			borrowed greatly from A68util.c
			fixes as late as April 2010. Also see "a18octal.txt" to
			change listing from hexidecimal values to octal.
	  		Fixes for Turbo C HRJ include line[] changed to lline[]

	JAN 2014	HRJ changed from A85 to A08 for 8008 assembly
	            1) changed bccsearc() to simple linear search
				2) add DATA pseudoop, like DB
	May 2014	HRJ split octal address in listings

	Aug 2014	HRJ changed for 8080-like mnemonics


This module contains the following utility packages:

	1)  symbol table building and searching

	2)  opcode and operator table searching

	3)  listing file output

	4)  hex file output

	5)  error flagging
*/

/*  Get global goodies:  */

#include "A8008.h"
#include <string.h> /* HRJ */
#include <ctype.h>
// #include <malloc.h> /* for lcc-32 HRJ */
/* #include <alloc.h> for Turbo C HRJ */
#include <stdlib.h>

/*  Make sure that MSDOS compilers using the large memory model know	*/
/*  that calloc() returns pointer to char as an MSDOS far pointer is	*/
/*  NOT compatible with the int type as is usually the case.		*/

/* char *calloc(); HRJ */

/*HRJ local declarations */

static OPCODE *bccsearch(OPCODE *, OPCODE *, char *);
static void list_sym(SYMBOL *);
static void record(unsigned);
static void putb(unsigned);
static int ustrcmp(char *, char*);
static void check_page(void);
void warning(char *);
void fatal_error(char *);


/*  Get access to global mailboxes defined in A8008.C:			*/

extern char errcode, lline[], title[];
extern int eject, listhex;
extern unsigned address, bytes, errors, listleft, obj[], pagelen, split_val;

/*  The symbol table is a binary tree of variable-length blocks drawn	*/
/*  from the heap with the calloc() function.  The root pointer lives	*/
/*  here:								*/

static SYMBOL *sroot = NULL;

/*  Add new symbol to symbol table.  Returns pointer to symbol even if	*/
/*  the symbol already exists.  If there's not enough memory to store	*/
/*  the new symbol, a fatal error occurs.				*/

SYMBOL *new_symbol(char *nam)

{
    SCRATCH int i;
    SCRATCH SYMBOL **p, *q;
    void fatal_error(char *);

    /* printf("new_symbol>>%s<<\n",nam);  HRJ diagnostic*/

    for (p = &sroot; (q = *p) && (i = strcmp(nam,q -> sname)); )
	p = i < 0 ? &(q -> left) : &(q -> right);
    if (!q) {
	if (!(*p = q = (SYMBOL *)calloc(1,sizeof(SYMBOL) + strlen(nam))))
	    fatal_error(SYMBOLS);
	strcpy(q -> sname,nam);
    }
    return q;
}

/*  Look up symbol in symbol table.  Returns pointer to symbol or NULL	*/
/*  if symbol not found.						*/

SYMBOL *find_symbol(char *nam)

{
    SCRATCH int i;
    SCRATCH SYMBOL *p;

    for (p = sroot; p && (i = strcmp(nam,p -> sname));
	p = i < 0 ? p -> left : p -> right);
    return p;
}

/*  Opcode table search routine.  This routine pats down the opcode	*/
/*  table for a given opcode and returns either a pointer to it or	*/
/*  NULL if the opcode doesn't exist.					*/
/*  HRJ table search assumes alpha order on name string field!! */

OPCODE *find_code(char *nam)

{
    /* OPCODE *bsearch(); */

    static OPCODE opctbl[] = {


	{ PSEUDO + ISIF,		ELSE,		"ELSE"	}, // no change HRJ
	{ PSEUDO + ISIF,		ENDIF,	"ENDIF"	},
	{ PSEUDO + ISIF,		IF,		"IF"	},
	{ PSEUDO,			DB,		"DB"	},
	{ PSEUDO,			DB,		"DATA"	},
	{ PSEUDO,			DS,		"DS"	},
	{ PSEUDO,			DW,		"DW"	},
	{ PSEUDO,			END,		"END"	},
	{ PSEUDO,			EQU,		"EQU"	},
	{ PSEUDO,			INCL,		"INCL"	},
	{ PSEUDO,			ORG,		"ORG"	},
	{ PSEUDO,			PAGE,		"PAGE"	},
	{ PSEUDO,			SET,		"SET"	},
	{ PSEUDO,			TITLE,	"TITLE"	},
	{ PSEUDO,          	SPLIT,  "SPLIT" },  //HRJ for split octal

	//HRJ replaced with 8080-version 8008 opcodes

	{ DATA_8 + 2,				0x04,	"ADI"	}, //HRJ immediates
	{ DATA_8 + 2,				0x0c,	"ACI"	},
	{ DATA_8 + 2,				0x14,	"SUI"	},
	{ DATA_8 + 2,				0x1C,	"SBI"	},
	{ DATA_8 + 2,				0x24,	"ANI"	},
	{ DATA_8 + 2,				0x2C,	"XRI"	},
	{ DATA_8 + 2,				0x34,	"ORI"	},
	{ DATA_8 + 2,				0x3C,	"CPI"	},


	{ DATA_16 + 3,				0x44,	"JMP"	}, //HRJ jump
	{ DATA_16 + 3,				0x40,	"JNC"	},//HRJ cond jumps
	{ DATA_16 + 3,				0x48,	"JNZ"	},
	{ DATA_16 + 3,				0x50,	"JP"	},
	{ DATA_16 + 3,				0x58,	"JPO"	},
	{ DATA_16 + 3,				0x60,	"JC"	},
	{ DATA_16 + 3,				0x68,	"JZ"	},
	{ DATA_16 + 3,				0x70,	"JM"	},
	{ DATA_16 + 3,				0x78,	"JPE"	},


	{ DATA_16 + 3,				0x46,	"CALL"	}, //HRJ call

	{ DATA_16 + 3,				0x42,	"CNC"	},//HRJ cond calls
	{ DATA_16 + 3,				0x4A,	"CNZ"	},
	{ DATA_16 + 3,				0x52,	"CP"	},
	{ DATA_16 + 3,				0x5A,	"CPO"	},
	{ DATA_16 + 3,				0x62,	"CC"	},
	{ DATA_16 + 3,				0x6A,	"CZ"	},
	{ DATA_16 + 3,				0x72,	"CM"	},
	{ DATA_16 + 3,				0x7A,	"CPE"	},

	{ DAD_REG + (DATA_LXI << 4) + 3,		0x06,	"LXI"	},//HRJ fake LXI to two MVI
	//{ DAD_REG + 3,		0x06,	"LXI"	},//HRJ fake LXI to two MVI

	{ NONE + 1,				0x02,	"RLC"	}, //HRJ rotates
	{ NONE + 1,				0x0A,	"RRC"	},
	{ NONE + 1,				0x12,	"RAL"	},
	{ NONE + 1,				0x1A,	"RAR"	},

	{ NONE + 1,				0x07,	"RET"	}, //HRJ returns
	{ NONE + 1,				0x0B,	"RNZ"	},
	{ NONE + 1,				0x2B,	"RZ"	},
	{ NONE + 1,				0x03,	"RNC"	},
	{ NONE + 1,				0x23,	"RC"	},
	{ NONE + 1,				0x1B,	"RPO"	},
	{ NONE + 1,				0x3b,	"RPE"	},
	{ NONE + 1,				0x13,	"RP"	},
	{ NONE + 1,				0x33,	"RM"	},

	{ DST_REG + 1,				0x00,	"INR"	},  //HRJ incr and decr
	{ DST_REG + 1,				0x01,	"DCR"	},
	// HRJ INA DCA are not accepted 00, 10 are Intel "halts"

	{ SRC_REG + 1,			0x80,	"ADD"	}, //HRJ ADD s
	{ SRC_REG + 1,			0x88,	"ADC"	}, //ADD w carry
	{ SRC_REG + 1,			0x90,	"SUB"	}, // SUBtract
	{ SRC_REG + 1,			0x98,	"SBB"	}, // SUB w borrow
	{ SRC_REG + 1,			0xa0,	"ANA"	}, // AND
	{ SRC_REG + 1,			0xA8,	"XRA"	}, // XOR
	{ SRC_REG + 1,			0xB0,	"ORA"	}, //OR
	{ SRC_REG + 1,			0xB8,	"CMP"	}, // COMPARE

	{ NONE + 1,				0xC0,	"NOP"	}, // MOV B,B is NOP HRJ move registers

	{ DST_REG + (SRC_REG << 4) + 1,	0xC0,	"MOV"	}, // 11DDDSSS
	{ DST_REG + (DATA_8 << 4) + 2,	0x06,	"MVI"	}, // 00DDD110

// 8080 version of MOV D,S is 01DDDSSS so should be OK
// 8080 version of MVI D,value is 00DDD110 so should be OK

	{ NONE + 1,				0xFF,	"HLT"	}, // MOV M,M is halt

	{ PORT + 1,				0x51,	"OUT"	}, // out is 01RRMMM1, RR is 1, 2, 3 only
	{ PORT + 1,				0x41,	"IN"	}, // in  is 0100MMM1



	{ RST_NUM + 1,			0x05,	"RST"	} // HRJ
    };

    return bccsearch(opctbl,opctbl + (sizeof(opctbl) / sizeof(OPCODE)),nam);
}

/*  Operator table search routine.  This routine pats down the		*/
/*  operator table for a given operator and returns either a pointer	*/
/*  to it or NULL if the opcode doesn't exist.				*/

OPCODE *find_operator(char *nam)

{
    /* OPCODE *bsearch(); */

    static OPCODE oprtbl[] = {
	{ BCDEHLMA + REG,				A,	"A"	},
	{ BINARY + LOG1  + OPR,				AND,	"AND"	},
	{ BCDEHLMA + BDHPSW + BDHSP + BD + REG,		B,	"B"	},
	{ BCDEHLMA + REG,				C,	"C"	},
	{ BCDEHLMA + BDHPSW + BDHSP + BD + REG,		D,	"D"	},
	{ BCDEHLMA + REG,				E,	"E"	},
	{ BINARY + RELAT + OPR,				'=',	"EQ"	},
	{ BINARY + RELAT + OPR,				GE,	"GE"	},
	{ BINARY + RELAT + OPR,				'>',	"GT"	},
	{ BCDEHLMA + BDHPSW + BDHSP + REG,		H,	"H"	},
	{ UNARY  + UOP3  + OPR,				HIGH,	"HIGH"	},
	{ BCDEHLMA + REG,				L,	"L"	},
	{ BINARY + RELAT + OPR,				LE,	"LE"	},
	{ UNARY  + UOP3  + OPR,				LOW,	"LOW"	},
	{ BINARY + RELAT + OPR,				'<',	"LT"	},
	{ BCDEHLMA + REG,				M,	"M"	},
	{ BINARY + MULT  + OPR,				MOD,	"MOD"	},
	{ BINARY + RELAT + OPR,				NE,	"NE"	},
	{ UNARY  + UOP2  + OPR,				NOT,	"NOT"	},
	{ BINARY + LOG2  + OPR,				OR,	"OR"	},
	{ BDHPSW + REG,					PSW,	"PSW"	},
	{ BINARY + MULT  + OPR,				SHL,	"SHL"	},
	{ BINARY + MULT  + OPR,				SHR,	"SHR"	},
	{ BDHSP + REG,					SP,	"SP"	},
	{ BINARY + LOG2  + OPR,				XOR,	"XOR"	}
    };

    return bccsearch(oprtbl,oprtbl + (sizeof(oprtbl) / sizeof(OPCODE)),nam);
}

// static OPCODE *bccsearch(OPCODE *lo, OPCODE *hi, char *nam)
//
// {
//    SCRATCH int i;
//    SCRATCH OPCODE *chk;
//
//    for (;;) {
//	chk = lo + (hi - lo) / 2;
//	if (!(i = ustrcmp(chk -> oname,nam))) return chk;
//	if (chk == lo) return NULL;
//	if (i < 0) lo = chk;
//	else hi = chk;
//    }
// }

static OPCODE *bccsearch(OPCODE *lo, OPCODE *hi, char *nam)
//HRJ simple search to avoid sorting op code table.

{
	SCRATCH int i;
	SCRATCH OPCODE *chk;

	for (chk = lo; chk < hi; chk++) {
	   if (!(i = ustrcmp(chk -> oname,nam))) return chk;
	}
	return NULL;
}

static int ustrcmp(char *s, char *t)

{
    SCRATCH int i;

    while (!(i = toupper(*s++) - toupper(*t)) && *t++);
    return i;
}

/*  Buffer storage for line listing routine.  This allows the listing	*/
/*  output routines to do all operations without the main routine	*/
/*  having to fool with it.						*/

static FILE *list = NULL;

/*  Listing file open routine.  If a listing file is already open, a	*/
/*  warning occurs.  If the listing file doesn't open correctly, a	*/
/*  fatal error occurs.  If no listing file is open, all calls to	*/
/*  lputs() and lclose() have no effect.				*/

void lopen(char *nam)

{

    if (list) warning(TWOLST);
    else if (!(list = fopen(nam,"w"))) fatal_error(LSTOPEN);
    return;
}

/*  Listing file line output routine.  This routine processes the	*/
/*  source line saved by popc() and the output of the line assembler in	*/
/*  buffer obj into a line of the listing.  If the disk fills up, a	*/
/*  fatal error occurs.							*/

/* HRJ change to OCTAL listing from hex */

void lputs(void)
{
    SCRATCH int i, j;
    SCRATCH unsigned *o;
    void fatal_error(char *);
	int address_hi, address_lo;

    if (list) {
	i = bytes;  o = obj;
	do {
	    fprintf(list,"%c  ",errcode);
	    if (listhex) {
		// fprintf(list,"%06o  ",address); non split octal
		address_hi = address / 256; address_lo = address - (address_hi * 256);
		fprintf(list,"%03o %03o  ",address_hi, address_lo);// split octal address
		for (j = 4; j; --j) {
		    if (i) { --i;  ++address;  fprintf(list," %03o",*o++); }
		    else fprintf(list,"    ");
		}
	    }
	    else fprintf(list,"%24s","");
	    fprintf(list,"   %s",lline);  strcpy(lline,"\n");
	    check_page();
	    if (ferror(list)) fatal_error(DSKFULL);
	} while (listhex && i);
    }
    return;
}

/*  Listing file close routine.  The symbol table is appended to the	*/
/*  listing in alphabetic order by symbol name, and the listing file is	*/
/*  closed.  If the disk fills up, a fatal error occurs.		*/

static int col = 0;

void lerror(void)
{
	 if (errors && list) fprintf(list, "%d Error(s)\n",errors); //hrj
}

void lclose(void)
{

    if (list) {
	if (sroot) {
	    list_sym(sroot);
	    if (col) fprintf(list,"\n");
	}
	fprintf(list,"\f");
	if (ferror(list) || fclose(list) == EOF) fatal_error(DSKFULL);
    }
    return;
}

static void list_sym(SYMBOL *sp)
{
	int valu_lo, valu_hi;

 	if (sp) {
	list_sym(sp -> left);
	// fprintf(list,"%04x  %-10s",sp -> valu,sp -> sname);
	valu_hi = sp -> valu / 256; valu_lo = sp -> valu - (valu_hi * 256);
	fprintf(list,"%03o %03o  %-10s",valu_hi, valu_lo, sp -> sname);  //HRJ octal
	if (col = ++col % SYMCOLS) fprintf(list,"    ");
	else {
	    fprintf(list,"\n");
	    if (sp -> right) check_page();
	}
	list_sym(sp -> right);
    }
    return;
}

static void check_page(void)
{
    if (pagelen && !--listleft) eject = TRUE;
    if (eject) {
	eject = FALSE;  listleft = pagelen;  fprintf(list,"\f");
	if (title[0]) { listleft -= 2;  fprintf(list,"%s\n\n",title); }
    }
    return;
}

/*  Buffer storage for hex output file.  This allows the hex file	*/
/*  output routines to do all of the required buffering and record	*/
/*  forming without the	main routine having to fool with it.		*/

static FILE *hex = NULL;
static unsigned cnt = 0;
static unsigned addr = 0;
static unsigned sum = 0;
static unsigned buf[HEXSIZE];

/*  Hex file open routine.  If a hex file is already open, a warning	*/
/*  occurs.  If the hex file doesn't open correctly, a fatal error	*/
/*  occurs.  If no hex file is open, all calls to hputc(), hseek(), and	*/
/*  hclose() have no effect.						*/

void hopen(char *nam)

{

    if (hex) warning(TWOHEX);
    else if (!(hex = fopen(nam,"w"))) fatal_error(HEXOPEN);
    return;
}

/*  Hex file write routine.  The data byte is appended to the current	*/
/*  record.  If the record fills up, it gets written to disk.  If the	*/
/*  disk fills up, a fatal error occurs.				*/

void hputc(unsigned c) // from hputc() HRJ

{

    if (hex) {
	buf[cnt++] = c;
	if (cnt == HEXSIZE) record(0);
    }
    return;
}

/*  Hex file address set routine.  The specified address becomes the	*/
/*  load address of the next record.  If a record is currently open,	*/
/*  it gets written to disk.  If the disk fills up, a fatal error	*/
/*  occurs.								*/

void hseek(unsigned a)

{

    if (hex) {
	if (cnt) record(0);
	addr = a;
    }
    return;
}

/*  Hex file close routine.  Any open record is written to disk, the	*/
/*  EOF record is added, and file is closed.  If the disk fills up, a	*/
/*  fatal error occurs.							*/

void hclose(void)
{

    if (hex) {
	   if (cnt) record(0);
	   record(1);
	   if (fclose(hex) == EOF) fatal_error(DSKFULL);
    }
    return;
}

static void record(unsigned typ)

{
    SCRATCH unsigned i;

	putc(':',hex);  putb(cnt);  putb(high(addr));
    putb(low(addr));  putb(typ);
    for (i = 0; i < cnt; ++i) putb(buf[i]);
    putb(low(0-sum));  putc('\n',hex); /* was (-sum) HRJ*/

    addr += cnt;  cnt = 0;

    if (ferror(hex)) fatal_error(DSKFULL);
    return;
}

static void putb(unsigned b)

{
    static char digit[] = "0123456789ABCDEF";

    putc(digit[b >> 4],hex);  putc(digit[b & 0x0f],hex);
    sum += b;  return;
}

/*  Error handler routine.  If the current error code is non-blank,	*/
/*  the error code is filled in and the	number of lines with errors	*/
/*  is adjusted.							*/

void error(char code)

{
    if (errcode == ' ') { errcode = code;  ++errors; }
    return;
}

/*  Fatal error handler routine.  A message gets printed on the stderr	*/
/*  device, and the program bombs.					*/

void fatal_error(char *msg)

{
    printf("Fatal Error -- %s\n",msg);
    exit(-1);
}

/*  Non-fatal error handler routine.  A message gets printed on the	*/
/*  stderr device, and the routine returns.				*/

void warning(char *msg)

{
    printf("Warning -- %s\n",msg);
    return;
}

